#![no_std]
#![no_main]

use panic_halt as _;

use longan_nano::hal::{
    pac,
    pac::ECLIC,
    prelude::*,
    gpio,
    timer::{Timer, Event},
    eclic::{EclicExt, LevelPriorityBits, TriggerType, Level, Priority},
};
use longan_nano::{lcd, lcd_pins};
use riscv_rt::entry;
use embedded_hal::digital::v2::{InputPin, OutputPin};

use dtlogic::{
    ui,
    input,
};

static mut G_R: Option<gpio::gpioc::PC13<gpio::Output<gpio::PushPull>>> = None;
static mut G_TIMER1: Option<Timer<pac::TIMER1>> = None;

#[entry]
fn main() -> ! {
    let dp = pac::Peripherals::take().unwrap();

    // Configure clocks
    let mut rcu = dp
        .RCU
        .configure()
        .ext_hf_clock(8.mhz())
        .sysclk(108.mhz())
        .freeze();
    let mut afio = dp.AFIO.constrain(&mut rcu);

    // Configure GPIO.
    let gpioa = dp.GPIOA.split(&mut rcu);
    let gpiob = dp.GPIOB.split(&mut rcu);
    let gpioc = dp.GPIOC.split(&mut rcu);

    // Blink.
    let (mut r, mut g, mut b) = (
        gpioc.pc13.into_push_pull_output(),
        gpioa.pa1.into_push_pull_output(),
        gpioa.pa2.into_push_pull_output(),
    );
    r.set_high().unwrap();
    g.set_high().unwrap();
    b.set_high().unwrap();

    // Relay output.
    let mut relay = gpiob.pb9.into_push_pull_output();
    relay.set_low().unwrap();

    // Set up LCD.
    let lcd_pins = lcd_pins!(gpioa, gpiob);
    let mut lcd = lcd::configure(dp.SPI0, lcd_pins, &mut afio, &mut rcu);

    // Configure interrupts and timer.
    ECLIC::reset();
    ECLIC::set_threshold_level(Level::L0);
    ECLIC::set_level_priority_bits(LevelPriorityBits::L3P1);
    ECLIC::setup(pac::Interrupt::TIMER1, TriggerType::Level, Level::L0, Priority::P0);
    unsafe { ECLIC::unmask(pac::Interrupt::TIMER1) };

    let mut timer = Timer::timer1(dp.TIMER1, 1.khz(), &mut rcu);
    timer.listen(Event::Update);

    unsafe {
        G_R = Some(r);
        G_TIMER1 = Some(timer);
    }

    unsafe { riscv::interrupt::enable() };

    // Run application code..
    
    let debug = false;

    let buttons = [
        gpioa.pa3.into_pull_up_input().downgrade(),
        gpioa.pa4.into_pull_up_input().downgrade(),
        gpiob.pb10.into_pull_up_input().downgrade(),
        gpiob.pb11.into_pull_up_input().downgrade(),
    ];

    let mut state_ts = ui::TimeSelect::new();
    let mut state_running = ui::Running::new();
    let mut s: &mut dyn ui::State<_> = &mut state_ts;
    s.enter(ui::Transition::TimeSelect);

    let mut ctx = ui::Context{
        buttons: input::Buttons::new(),
        ticks: 0,
    };

    loop {
        let mut states = [false; 4];
        for (i, button) in buttons.iter().enumerate() {
            if let Ok(v) = button.is_low() {
                states[i] = v;
            }
        }
        if debug {
            if states[0] {
                g.set_low().unwrap();
            } else {
                g.set_high().unwrap();
            }
            if states[3] {
                b.set_low().unwrap();
            } else {
                b.set_high().unwrap();
            }
        }
        ctx.ticks = ticks();
        ctx.buttons.update(ctx.ticks, states);
        let next = s.tick(&ctx);
        s.draw(&mut lcd).unwrap();
        match next {
            ui::Transition::Keep => (),
            ui::Transition::Running { .. } => {
                s = &mut state_running;
                s.enter(next);
                // Jank.
                relay.set_high().unwrap();
            },
            ui::Transition::TimeSelect => {
                s = &mut state_ts;
                s.enter(next);
                // Jank.
                relay.set_low().unwrap();
            },
        }
    }
}

static mut TICKS: u32 = 0;

fn ticks() -> u32 {
    let mut res: u32 = 0;
    riscv::interrupt::free(|_| {
        res = unsafe { TICKS };
    });
    res
}

#[allow(non_snake_case)]
#[no_mangle]
fn TIMER1() {
    unsafe {
        if let Some(timer) = &mut G_TIMER1 {
            timer.clear_update_interrupt_flag();
        }
        TICKS += 1;
    }
}
